/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import org.apache.commons.io.IOUtils;
import org.junit.Test;
import org.ow2.petals.cli.api.command.AbstractCommand;
import org.ow2.petals.cli.api.command.Command;
import org.ow2.petals.cli.api.command.exception.CommandBadArgumentNumberException;
import org.ow2.petals.cli.api.command.exception.CommandException;
import org.ow2.petals.cli.api.command.exception.CommandMissingArgumentException;
import org.ow2.petals.cli.api.command.exception.ConnectionErrorException;
import org.ow2.petals.cli.base.junit.AbstractConsoleInOutTest;
import org.ow2.petals.cli.command.CommandNotRequiringConnection;
import org.ow2.petals.cli.command.CommandRequiringConnection;
import org.ow2.petals.cli.command.CommandWithAnOptionalArgument;
import org.ow2.petals.cli.command.CommandWithSameArgumentThanConnect;
import org.ow2.petals.cli.command.Connect;
import org.ow2.petals.cli.command.Help;
import org.ow2.petals.cli.shell.ShellFactory;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit test of {@link AbstractMain}.
 * 
 * @author Christophe DENEUX - Linagora
 */
public class AbstractMainTest extends AbstractConsoleInOutTest {

    private static final String USAGE_HEADER = "The Usage Header of a Main for test purpose";

    private class MainForTestPurpose extends AbstractMain {

        public MainForTestPurpose() {
            super(new Connect());
        }

        @Override
        protected String getUsageHeader() {
            return USAGE_HEADER;
        }

        @Override
        protected String getPromptBase() {
            return null;
        }

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>no operand flag,</li>
     * <li>no command,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error because of the missing command,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of Petals CLI is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_00() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.", error.contains(ShellFactory.MISSING_COMMAND));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertTrue("Usage header is missing", output.contains(USAGE_HEADER));
        assertTrue("Usage footer is missing", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>the operand flag,</li>
     * <li>no command,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c --</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error because of the missing command,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of Petals CLI is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_01() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--" }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.", error.contains(ShellFactory.MISSING_COMMAND));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertTrue("Usage header is missing", output.contains(USAGE_HEADER));
        assertTrue("Usage footer is missing", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>a command needing no argument,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c -- help</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error because connection arguments are not required,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of Petals CLI is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_02() {

        final AbstractMain main = new MainForTestPurpose();
        final AbstractCommand command = new Help();

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", command.getName() },
                null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());
        assertFalse("Usage header is print", output.contains(USAGE_HEADER));
        assertFalse("Usage footer is print", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>a command containing an unrecognized option,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c -- list -q unrecognizedOption</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error because of the unrecognized argument of the command,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of Petals CLI is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_03() {

        final AbstractMain main = new MainForTestPurpose();

        final AbstractCommand command = new CommandWithAnOptionalArgument();
        final String unrecognizedOption = "q";

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", command.getName(),
                "-" + unrecognizedOption, "unrecognizedOption" }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertTrue("Command usage is missing", error.contains(command.getUsage()));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode with connection arguments,</li>
     * <li>a command containing an unrecognized option,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -h localhost -n 7700 -u petals -p petals -c -- list -q unrecognizedOption</code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error about the command because of unrecognized argument,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of the command is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_04() {

        final AbstractMain main = new MainForTestPurpose();

        final AbstractCommand command = new CommandWithAnOptionalArgument();
        final String unrecognizedOption = "q";

        final int exitCode = main.run(
                new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", "-" + ShellFactory.COMMAND_SHORT_OPTION,
                        "--", command.getName(), "-" + unrecognizedOption, "unrecognizedOption" },
                null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertTrue("Bad arguments error not detected",
                error.contains(CommandBadArgumentNumberException.BAD_NUMBER_OF_ARGUMENTS_OR_INCOMPATIBLE_ARGUMENTS));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertTrue("The command usage is missing in error message", error.contains(command.getUsage()));


    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode with connection arguments,</li>
     * <li>a command containing a missing argument for an option,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -h localhost -n 7700 -u petals -p petals -c -- list -t</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error because of the missing argument of the option,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of the command is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_05() {

        final AbstractMain main = new MainForTestPurpose();

        final AbstractCommand command = new CommandWithAnOptionalArgument();

        final int exitCode = main.run(
                new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", "-" + ShellFactory.COMMAND_SHORT_OPTION,
                        "--", command.getName(), "-" + CommandWithAnOptionalArgument.OPTIONAL_SHORT_OPTION },
                null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertFalse("The exception thrown is present in error message",
                error.contains(CommandMissingArgumentException.class.getName()));
        assertTrue("The error is not the expected one", error.contains(CommandMissingArgumentException.MISSING_ARGUMENT
                + ": " + CommandWithAnOptionalArgument.OPTIONAL_SHORT_OPTION));
        assertTrue("Command usage is missing", error.contains(command.getUsage()));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertFalse("Petals CLI usage header is present", output.contains(USAGE_HEADER));
        assertFalse("Petals CLI usage footer is present", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is enabled,</li>
     * <li>the command mode with connection arguments,</li>
     * <li>a command containing a missing argument for an option,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -h localhost -n 7700 -u petals -p petals -d -c -- list -t</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error because of the missing argument of the option,</li>
     * <li>no stack trace is printed,</li>
     * <li>usage of the command is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_06() {

        final AbstractMain main = new MainForTestPurpose();

        final AbstractCommand command = new CommandWithAnOptionalArgument();

        final int exitCode = main.run(new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", "-d",
                "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", command.getName(),
                "-" + CommandWithAnOptionalArgument.OPTIONAL_SHORT_OPTION }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 1, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertFalse("The exception thrown is present in error message",
                error.contains(CommandMissingArgumentException.class.getName()));
        assertTrue("The error is not the expected one", error.contains(CommandMissingArgumentException.MISSING_ARGUMENT
                + ": " + CommandWithAnOptionalArgument.OPTIONAL_SHORT_OPTION));
        assertTrue("Command usage is missing", error.contains(command.getUsage()));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertFalse("Petals CLI usage header is present", output.contains(USAGE_HEADER));
        assertFalse("Petals CLI usage footer is present", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Use case of <a href="http://jira.petalslink.com/browse/PETALSESBCLI-41">PETALSCLI-41</a>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * 
     * @see <a href="http://jira.petalslink.com/browse/PETALSESBCLI-41">PETALSCLI-41</a>
     */
    @Test
    public final void testRun_Command_PETALSCLI_41() throws IOException, CommandException {


        final AbstractMain main = new MainForTestPurpose();

        final String mandatoryArgValue = "mandatory-arg-value";
        final int exitCode = main.run(
                new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", "-" + ShellFactory.COMMAND_SHORT_OPTION,
                        "--", CommandWithSameArgumentThanConnect.NAME,
                        "-" + CommandWithSameArgumentThanConnect.MANDATORY_SHORT_OPTION, mandatoryArgValue },
                null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("An error message print", error.isEmpty());
        assertTrue("Missing command output",
                systemOut.getLog().contains(CommandWithSameArgumentThanConnect.SUCCESSFUL_OUTPUT));


    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is enabled,</li>
     * <li>the command mode with connection arguments to fail the connection establishement,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -h {@value ContainerAdministrationMock#UNKNOWN_HOST} -n 7700 -u petals -p petals -d -c -- list</code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection, but raises an exception on the host
     * {@value ContainerAdministrationMock#UNKNOWN_HOST}.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>exit code 3 because of error during connection,</li>
     * <li>error about the connection to the host,</li>
     * <li>error stacktrace is printed,</li>
     * <li>no usage is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_Error_07() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + Connect.HOST_SHORT_OPTION, Connect.HOST_FAILING_CONNECTION,
                "-d", "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", CommandRequiringConnection.NAME }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 3, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("The exception throwing is missing in error message",
                error.contains(Connect.ERROR_MSG_FAILING_CONNECTION));
        assertTrue("Stack trace is missing in error message", error.contains(AbstractMain.class.getName()));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>the inlined mode,</li>
     * <li>the command to connect to an existing Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > ./petals-cli.sh - << EOF<br/>
     * connect -h localhost -n 7700 -u petals -p petals<br/>
     * list
     * EOF
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>no error stacktrace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>commands correctly executed.</li>
     * </ul>
     * </p>
     * 
     * @throws IOException
     */
    @Test
    public final void testRun_Inlined_00() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        systemIn.writeLine(Connect.NAME + " -" + Connect.HOST_SHORT_OPTION + " localhost");
        systemIn.writeLine(CommandRequiringConnection.NAME);
        systemIn.writeLine("exit");

        final int exitCode = main.run(new String[] { "-" }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());
        assertTrue("Missing command output", systemOut.getLog().contains(Connect.SUCCESSFUL_OUTPUT));
        assertTrue("Missing command output", systemOut.getLog().contains(CommandRequiringConnection.SUCCESSFUL_OUTPUT));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>connection defines on the command line,</li>
     * <li>the inlined mode,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > ./petals-cli.sh -h localhost -n 7700 -u petals -p petals - << EOF<br/>
     * list
     * EOF
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error occurs because the command 'connect' is not used before the command requiring a connection,</li>
     * <li>no error stacktrace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>command flow is interrupted on command in error.</li>
     * </ul>
     * </p>
     * 
     * @throws IOException
     */
    @Test
    public final void testRun_Inlined_Error_02() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        systemIn.writeLine(CommandRequiringConnection.NAME);

        final int exitCode = main.run(new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", "-" }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 2, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("Error not associated to the connection not established: " + error,
                error.contains("A connection must be established previously to execute the command"));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>the inlined mode,</li>
     * <li>the command to connect to an unknown Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > ./petals-cli.sh - << EOF<br/>
     * connect -h {@value ContainerAdministrationMock#UNKNOWN_HOST} -n 7700 -u petals -p petals<br/>
     * list
     * EOF
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection, but raises an exception on the host
     * {@value ContainerAdministrationMock#UNKNOWN_HOST}.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error on the command 'connect' about the connection to an unknown host,</li>
     * <li>no error stacktrace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>command flow is interrupted on command in error.</li>
     * </ul>
     * </p>
     * 
     * @throws IOException
     */
    @Test
    public final void testRun_Inlined_Error_00() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        systemIn.writeLine(Connect.NAME + " -" + Connect.HOST_SHORT_OPTION + " " + Connect.HOST_FAILING_CONNECTION);
        systemIn.writeLine(CommandRequiringConnection.NAME);

        final int exitCode = main.run(new String[] { "-" }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 3, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("Error not associated to the unknown host: " + error, error.contains("Unknown Host"));
        assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is enabled,</li>
     * <li>the inlined mode,</li>
     * <li>the command to connect to an unknown Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > ./petals-cli.sh -d - << EOF<br/>
     * connect -h {@value ContainerAdministrationMock#UNKNOWN_HOST} -n 7700 -u petals -p petals<br/>
     * list
     * EOF
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection, but raises an exception on the host '
     * {@value ContainerAdministrationMock#UNKNOWN_HOST} '.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error on the command 'connect' about the connection to an unknown host,</li>
     * <li>error stacktrace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>command flow is interrupted on command in error.</li>
     * </ul>
     * </p>
     * 
     * @throws IOException
     */
    @Test
    public final void testRun_Inlined_Error_01() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        systemIn.writeLine(Connect.NAME + " -" + Connect.HOST_SHORT_OPTION + " " + Connect.HOST_FAILING_CONNECTION);
        systemIn.writeLine(CommandRequiringConnection.NAME);

        final int exitCode = main.run(new String[] { "-d", "-" }, null, null);

        final String error = systemErr.getLog();
        assertEquals("invalid exit code", 3, exitCode);
        assertFalse("No error message print", error.isEmpty());
        assertTrue("The exception thrown is missing in error message: " + error,
                error.contains(ConnectionErrorException.class.getName()));
        assertTrue("Stack trace is missing in error message", error.contains("Caused by"));
        assertTrue("Stack trace is missing in error message", error.contains(AbstractMainTest.class.getName()));

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>the file mode,</li>
     * <li>the command to connect to an existing Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > cat petals-file.cmd<br/>
     * connect -h localhost -n 7700 -u petals -p petals<br/>
     * list<br/>
     * > ./petals-cli.sh petals-file.cmd
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>no error stacktrace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>commands correctly executed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_File_00() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        final File commandFile = File.createTempFile("petals-cli", "cmd");
        try {
            final OutputStream os = new FileOutputStream(commandFile);
            final Writer sw = new OutputStreamWriter(os);
            try {
                final StringBuilder commandFlow = new StringBuilder();
                commandFlow.append(Connect.NAME).append(' ').append("-" + Connect.HOST_SHORT_OPTION).append(' ')
                        .append("localhost").append(System.getProperty("line.separator"));
                commandFlow.append(CommandRequiringConnection.NAME).append(System.getProperty("line.separator"));
                sw.write(commandFlow.toString());
            } finally {
                sw.close();
            }

            final int exitCode = main.run(new String[] { commandFile.getAbsolutePath() }, null, null);

            final String error = systemErr.getLog();
            assertEquals("invalid exit code", 0, exitCode);
            assertTrue("Error message print", error.isEmpty());
            assertTrue("Missing command output", systemOut.getLog().contains(Connect.SUCCESSFUL_OUTPUT));
            assertTrue("Missing command output",
                    systemOut.getLog().contains(CommandRequiringConnection.SUCCESSFUL_OUTPUT));

        } finally {
            commandFile.delete();
        }
    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>connection defines on the command line,</li>
     * <li>the file mode,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > cat petals-file.cmd<br/>
     * list<br/>
     * > ./petals-cli.sh -h localhost -n 7700 -u petals -p petals petals-file.cmd
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>an error occurs about the connection not established,</li>
     * <li>no error stacktrace is printed,</li>
     * <li>no usage is printed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_File_Error_02() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        final File commandFile = File.createTempFile("petals-cli", "cmd");
        try {
            final OutputStream os = new FileOutputStream(commandFile);
            final Writer sw = new OutputStreamWriter(os);
            try {
                final StringBuilder commandFlow = new StringBuilder(CommandRequiringConnection.NAME)
                        .append(System.getProperty("line.separator"));
                sw.write(commandFlow.toString());
            } finally {
                sw.close();
            }

            final int exitCode = main.run(
                    new String[] { "-" + Connect.HOST_SHORT_OPTION, "localhost", commandFile.getAbsolutePath() }, null,
                    null);

            final String error = systemErr.getLog();
            assertEquals("invalid exit code", 2, exitCode);
            assertFalse("No error message print", error.isEmpty());
            assertTrue("Error not associated to the connection not established: " + error,
                    error.contains("A connection must be established previously to execute the command"));
            assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
            assertFalse("Stack trace is present in error message", error.contains("Caused by"));

        } finally {
            commandFile.delete();
        }

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is disable,</li>
     * <li>the file mode,</li>
     * <li>the command to connect to an unknown Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > cat petals-file.cmd<br/>
     * connect -h unknown -n 7700 -u petals -p petals<br/>
     * list<br/>
     * > ./petals-cli.sh petals-file.cmd
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection, but raises an exception on the host 'unknown'.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error on the command 'connect' about the connection to an unknown host,</li>
     * <li>no error stack trace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>command flow is interrupted on command in error.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_File_Error_00() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        final File commandFile = File.createTempFile("petals-cli", "cmd");
        try {
            final OutputStream os = new FileOutputStream(commandFile);
            final Writer sw = new OutputStreamWriter(os);
            try {
                final StringBuilder commandFlow = new StringBuilder();
                commandFlow.append(Connect.NAME).append(' ').append("-" + Connect.HOST_SHORT_OPTION).append(' ')
                        .append(Connect.HOST_FAILING_CONNECTION).append(System.getProperty("line.separator"));
                commandFlow.append(CommandRequiringConnection.NAME).append(System.getProperty("line.separator"));
                sw.write(commandFlow.toString());
            } finally {
                sw.close();
            }

            final int exitCode = main.run(new String[] { commandFile.getAbsolutePath() }, null, null);

            final String error = systemErr.getLog();
            assertEquals("invalid exit code", 3, exitCode);
            assertFalse("No error message print", error.isEmpty());
            assertTrue("Error not associated to the unknown host: " + error, error.contains("Unknown Host"));
            assertFalse("Stack trace is present in error message", error.contains(MainForTestPurpose.class.getName()));
            assertFalse("Stack trace is present in error message", error.contains("Caused by"));

        } finally {
            commandFile.delete();
        }

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the flag debug is enabled,</li>
     * <li>the file mode,</li>
     * <li>the command to connect to an unknown Petals node,</li>
     * <li>a command needing no argument and requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>
     * > cat petals-file.cmd<br/>
     * connect -h unknown -n 7700 -u petals -p petals<br/>
     * list<br/>
     * > ./petals-cli.sh -d petals-file.cmd
     * </code>
     * </p>
     * <p>
     * Note: the connection <b>is not a real</b> connection, but raises an exception on the host 'unknown'.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error on the command 'connect' about the connection to an unknown host,</li>
     * <li>error stack trace is printed,</li>
     * <li>no usage is printed,</li>
     * <li>command flow is interrupted on command in error.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_File_Error_01() throws IOException {

        final AbstractMain main = new MainForTestPurpose();

        final File commandFile = File.createTempFile("petals-cli", "cmd");
        try {
            final OutputStream os = new FileOutputStream(commandFile);
            final Writer sw = new OutputStreamWriter(os);
            try {
                final StringBuilder commandFlow = new StringBuilder();
                commandFlow.append(Connect.NAME).append(' ').append("-" + Connect.HOST_SHORT_OPTION).append(' ')
                        .append(Connect.HOST_FAILING_CONNECTION).append(System.getProperty("line.separator"));
                commandFlow.append(CommandRequiringConnection.NAME).append(System.getProperty("line.separator"));
                sw.write(commandFlow.toString());
            } finally {
                sw.close();
            }

            final int exitCode = main.run(new String[] { "-d", commandFile.getAbsolutePath() }, null, null);

            final String error = systemErr.getLog();
            assertEquals("invalid exit code", 3, exitCode);
            assertFalse("No error message print", error.isEmpty());
            assertTrue("The exception thrown is missing in error message: " + error,
                    error.contains(ConnectionErrorException.class.getName()));
            assertTrue("Stack trace is missing in error message", error.contains("Caused by"));
            assertTrue("Stack trace is missing in error message", error.contains(AbstractMainTest.class.getName()));

        } finally {
            commandFile.delete();
        }
    }

    /**
     * <p>
     * Unit test about the default logging system configuration.
     * </p>
     * <p>
     * Expected results: Template of error message is: <code>ERROR: xxxxxxxx</code>.
     * </p>
     */
    @Test
    public final void testLog_00() throws IOException, InvocationTargetException {

        final AbstractMain main = new MainForTestPurpose();
        ReflectionHelper.invokePrivateMethod(AbstractMain.class, main, "initLogging", null, null);

        final String errorMessage = "My error message";
        Logger.getLogger(this.getClass().getCanonicalName()).severe(errorMessage);
        final String warnMessage = "My warn message";
        Logger.getLogger(this.getClass().getCanonicalName()).warning(warnMessage);
        final String infoMessage = "My info message";
        Logger.getLogger(this.getClass().getCanonicalName()).info(infoMessage);

        final String error = systemErr.getLog();
        assertTrue("Unexpected format",
                error.startsWith("ERROR: " + errorMessage + System.getProperty("line.separator")));
        assertTrue("Warn level is not printed", error.contains(warnMessage));
        assertFalse("Info level is printed", error.contains(infoMessage));

    }

    /**
     * <p>
     * Unit test about a logging system configuration provided at runtime.
     * </p>
     * <p>
     * Expected results: the configuration provided is correctly used.
     * </p>
     */
    @Test
    public final void testLog_01() throws IOException, InvocationTargetException {

        final File logConfFile = File.createTempFile("petals-cli-log", ".properties");
        final OutputStream osLogConfFile = new FileOutputStream(logConfFile);
        final InputStream isLogConfFile = Thread.currentThread().getContextClassLoader()
                .getResourceAsStream("test-log-conf.properties");
        try {
            IOUtils.copy(isLogConfFile, osLogConfFile);
        } finally {
            osLogConfFile.close();
            isLogConfFile.close();
        }

        try {
            // The main class MUST be instanced after to have set the console streams !!
            final AbstractMain main = new MainForTestPurpose();
            System.setProperty("java.util.logging.config.file", logConfFile.getAbsolutePath());
            ReflectionHelper.invokePrivateMethod(AbstractMain.class, main, "initLogging", null, null);

            final String errorMessage = "My error message";
            Logger.getLogger(this.getClass().getCanonicalName()).severe(errorMessage);
            final String warnMessage = "My warn message";
            Logger.getLogger(this.getClass().getCanonicalName()).warning(warnMessage);
            final String infoMessage = "My info message";
            Logger.getLogger(this.getClass().getCanonicalName()).info(infoMessage);

            final String error = systemErr.getLog();
            assertTrue("Unexpected format",
                    error.startsWith("ERROR: " + errorMessage + System.getProperty("line.separator")));
            assertTrue("Warn level is not printed", error.contains(warnMessage));
            assertTrue("Info level is not printed", error.contains(infoMessage));
        } finally {
            logConfFile.delete();
        }

    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>the operand flag,</li>
     * <li>a command needing an operand flag completed with an argument,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c -- mycommand -- -z</code>
     * </p>
     * <p>
     * Expected results: no error
     * </p>
     */
    @Test
    public final void testRun_Command_00() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--",
                CommandNotRequiringConnection.NAME, "--", "-z" }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());
        assertFalse("Usage header is not print", output.contains(USAGE_HEADER));
        assertFalse("Usage footer is not print", output.contains(AbstractMain.USAGE_FOOTER));
        assertTrue("Missing command output",
                systemOut.getLog().contains(CommandNotRequiringConnection.SUCCESSFUL_OUTPUT));
    }

    /**
     * <p>
     * Check the main entry point against an command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>no operand flag,</li>
     * <li>a command needing an operand flag completed with an argument,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c mycommand -- -z</code>
     * </p>
     * <p>
     * Expected results: no error
     * </p>
     * 
     * @see PETALSESBCLI-88
     */
    @Test
    public final void testRun_Command_01() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION,
                CommandNotRequiringConnection.NAME, "--", "-z" }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());

    }

    /**
     * <p>
     * Check the main entry point against default connection parameters, a command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>the operand flag,</li>
     * <li>a command requiring a connection,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c -- version</code>
     * </p>
     * <p>
     * Expected results: no error
     * </p>
     */
    @Test
    public final void testDefaultConnection_00() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(
                new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", CommandRequiringConnection.NAME }, null,
                null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());
        assertFalse("Usage header is not print", output.contains(USAGE_HEADER));
        assertFalse("Usage footer is not print", output.contains(AbstractMain.USAGE_FOOTER));
        assertTrue("Missing command output", systemOut.getLog().contains(CommandRequiringConnection.SUCCESSFUL_OUTPUT));

    }

    /**
     * <p>
     * Check the main entry point against no connection establishment to execute
     * 'help' in command line mode, ie. a command line containing:
     * <ul>
     * <li>the command mode without connection arguments,</li>
     * <li>the operand flag,</li>
     * <li>the command 'help' without argument,</li>
     * </ul>
     * as following:<br/>
     * <code>> ./petals-cli.sh -c -- help</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>general help message is displayed,</li>
     * <li>no connection is established,</li>
     * <li>usage header of Petals CLI is not displayed,</li>
     * <li>usage footer of Petals CLI is not displayed.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void testRun_Command_NoConnectionRequiredForHelp() {

        final AbstractMain main = new MainForTestPurpose();

        final Command helpCommand = new Help();
        final int exitCode = main
                .run(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", helpCommand.getName() }, null, null);

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("Error message print", error.isEmpty());
        assertFalse("A connection was established", Connect.IS_CONNECTED);
        assertTrue("Unexpected help message", output.contains("Available commands:"));
        assertFalse("Usage header is not print", output.contains(USAGE_HEADER));
        assertFalse("Usage footer is not print", output.contains(AbstractMain.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the version information get using the argument '-V' as following:<br/>
     * <code>> ./petals-cli.sh -V</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>3 lines are returned,</li>
     * <li>the first line is the version of CLI,</li>
     * <li>exit code is 0,</li>
     * <li>no error occurs.</li>
     * </ul>
     * </p>
     */
    @Test
    public final void getVersionInformationOfCLI() {

        final AbstractMain main = new MainForTestPurpose();

        final int exitCode = main.run(new String[] { "-" + AbstractMain.VERSION_SHORT_OPTION }, null, null);

        final String output = systemOut.getLog();
        final String error = systemErr.getLog();

        assertEquals("invalid exit code", 0, exitCode);
        assertTrue("An error occurs", error.isEmpty());
        assertFalse("Empty output", output.isEmpty());
        final String[] lines = output.split("\n");
        assertEquals(3, lines.length);
        assertTrue(lines[0].startsWith(USAGE_HEADER));

    }
}
