/**
 * Copyright (c) 2014-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.command;

import java.net.UnknownHostException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.MissingArgumentException;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.UnrecognizedOptionException;
import org.ow2.petals.cli.api.command.AbstractCommand;
import org.ow2.petals.cli.api.command.ConnectionCommand;
import org.ow2.petals.cli.api.command.exception.CommandBadArgumentNumberException;
import org.ow2.petals.cli.api.command.exception.CommandException;
import org.ow2.petals.cli.api.command.exception.CommandInvalidArgumentException;
import org.ow2.petals.cli.api.command.exception.CommandInvalidException;
import org.ow2.petals.cli.api.command.exception.CommandMissingArgumentException;
import org.ow2.petals.cli.api.command.exception.CommandMissingOptionsException;
import org.ow2.petals.cli.api.command.exception.ConnectionErrorException;
import org.ow2.petals.cli.api.connection.AuthenticatedConnectionParameters;
import org.ow2.petals.cli.api.connection.ConnectionParameters;
import org.ow2.petals.cli.api.connection.exception.InvalidConnectionParameterException;
import org.ow2.petals.cli.api.pref.PreferenceFileException;

/**
 * A dummy connection command for unit tests
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class Connect extends AbstractCommand implements ConnectionCommand {

    public static final String NAME = "connect-command";

    public static final String SUCCESSFUL_OUTPUT = NAME + " is ok";

    public static boolean IS_CONNECTED = false;

    /**
     * A mandatory option
     */
    public static final String HOST_SHORT_OPTION = "h";

    public static final String HOST_FAILING_CONNECTION = "12.34.56.78";

    public static final String ERROR_MSG_FAILING_CONNECTION = "Connection fails (fake: Unknown Host)";

    private static final Option HOST_OPTION = Option.builder(HOST_SHORT_OPTION).hasArg(true).required(false).build();

    public Connect() {
        super(NAME);
    }

    @Override
    public void doExecute(final String[] args) throws CommandException {

        try {
            final CommandLineParser parser = new DefaultParser();
            final CommandLine cmd = parser.parse(this.getOptions(), args);

            final ConnectionParameters connectionParameters = this.parseConnectionParameters(cmd, true);
            this.connect(connectionParameters, false);

        } catch (final UnrecognizedOptionException e) {
            throw new CommandBadArgumentNumberException(this, e);
        } catch (final MissingArgumentException e) {
            throw new CommandMissingArgumentException(this, e.getOption(), e);
        } catch (final MissingOptionException e) {
            throw new CommandMissingOptionsException(this, e.getMissingOptions(), e);
        } catch (final ParseException e) {
            throw new CommandInvalidException(this, e);
        } catch (final PreferenceFileException e) {
            throw new CommandException(this, e);
        } catch (final InvalidConnectionParameterException e) {
            throw new CommandException(this, e);
        }
        this.getShell().getPrintStream().println(SUCCESSFUL_OUTPUT);
    }

    @Override
    public AuthenticatedConnectionParameters connect(final ConnectionParameters connectionParameters,
            final boolean isYesFlagEnabledAtCmdLevel) throws ConnectionErrorException, CommandException {
        if (connectionParameters != null && HOST_FAILING_CONNECTION.equals(connectionParameters.getHost())) {
            throw new ConnectionErrorException(this, new Exception(Connect.ERROR_MSG_FAILING_CONNECTION));
        } else {
            IS_CONNECTED = true;
            this.getShell().onConnectionEstablished();
            return null;
        }
    }

    @Override
    public ConnectionParameters parseConnectionParameters(final CommandLine cmd, final boolean useDefaultConnection)
            throws CommandException, PreferenceFileException, InvalidConnectionParameterException {

        if (cmd.hasOption(HOST_SHORT_OPTION)) {
            try {
                return new ConnectionParameters(cmd.getOptionValue(HOST_SHORT_OPTION), 1000);
            } catch (final UnknownHostException e) {
                throw new CommandInvalidArgumentException(this, HOST_OPTION, cmd.getOptionValue(HOST_SHORT_OPTION));
            }
        } else {
            return null;
        }
    }

    @Override
    public Options getOptions() {
        final Options options = new Options();
        options.addOption(HOST_OPTION);
        return options;
    }

    @Override
    public boolean isConnectionRequired() {
        return false;
    }
}
