/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.command;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayOutputStream;

import org.junit.Test;
import org.ow2.petals.cli.api.command.exception.CommandBadArgumentNumberException;
import org.ow2.petals.cli.api.command.exception.CommandException;
import org.ow2.petals.cli.api.shell.exception.ShellException;
import org.ow2.petals.cli.base.junit.shell.DummyShell;

/**
 *
 * @author "Mathieu CARROLLE - EBM WebSourcing"
 *
 */
public class SystemCallTest {

    /**
     * Test that the command usage is defined
     */
    @Test
    public void testUsage_0() {

        final SystemCall command = new SystemCall();
        command.setShell(new DummyShell());
        final String usage = command.getUsage();
        assertNotNull("Command usage null", usage);
        assertFalse("Command usage empty", usage.isEmpty());
        assertTrue("'usage' is missing at the begining of the command usage",
                usage.startsWith("usage"));
        assertTrue("The command name is missing in the command usage",
                usage.contains(command.getName()));
    }

    /**
     * Test of the command "system" against argument management, where:
     * <ul>
     * <li>no option is provided</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: an error occurs about bad argument number.
     * </p>
     * 
     * @see <a
     *      href="http://jira.petalslink.com/browse/PETALSESBCLI-74">PETALSESBCLI-74</a>
     */
    @Test(expected = CommandBadArgumentNumberException.class)
    public void testArgumentsError_0() throws CommandException {
        final SystemCall command = new SystemCall();
        command.setShell(new DummyShell());

        command.execute(new String[] {});
    }

    /**
     * Test of the command "system" where:
     * <ul>
     * <li>a system command writing something on output is invoked,</li>
     * <li>the system command requires one argument, and it is given as only one argument in a Petals CLI point of view</li>
     * </ul>
     * </p> Expected results:
     * <ul>
     * <li>No error occurs,</li>
     * <li>The expected output is correctly printed,</li>
     * <li>Exit code is 0.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testExecuteAsOnlyOneArgument() throws ShellException, CommandException {
        final ByteArrayOutputStream errBAOS = new ByteArrayOutputStream();
        final ByteArrayOutputStream outBAOS = new ByteArrayOutputStream();

        final String expected = "linux";
        final ShellManagingExitCode shell = new ShellManagingExitCode(outBAOS, errBAOS);
        final SystemCall command = new SystemCall();
        command.setShell(shell);

        command.execute( new String[]{ "echo " + expected });

        assertEquals(expected + System.getProperty("line.separator"), outBAOS.toString());
        assertTrue(errBAOS.toString().isEmpty());
        assertEquals(0, shell.getExitStatus());
    }

    /**
     * Test of the command "system" where:
     * <ul>
     * <li>a system command writing something on output is invoked,</li>
     * <li>the system command requires one argument, and they are given as separated arguments in a Petals CLI point of
     * view</li>
     * </ul>
     * </p> Expected results:
     * <ul>
     * <li>No error occurs,</li>
     * <li>The expected output is correctly printed,</li>
     * <li>Exit code is 0.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testExecuteAsSeveralArguments() throws ShellException, CommandException {
        final ByteArrayOutputStream errBAOS = new ByteArrayOutputStream();
        final ByteArrayOutputStream outBAOS = new ByteArrayOutputStream();

        final String expected = "linux";
        final ShellManagingExitCode shell = new ShellManagingExitCode(outBAOS, errBAOS);
        final SystemCall command = new SystemCall();
        command.setShell(shell);

        command.execute(new String[] { "echo", expected });

        assertEquals(expected + System.getProperty("line.separator"), outBAOS.toString());
        assertTrue(errBAOS.toString().isEmpty());
        assertEquals(0, shell.getExitStatus());
    }

    @Test
    public void testExecuteWithErrorExitStatus() throws ShellException, CommandException {

        final String cmd;
        final String osName = System.getProperty("os.name");
    	if( osName != null
    			&& osName.toLowerCase().contains( "win" ))
    		cmd = "dir .\\lol31";
    	else
    		cmd = "ls /linuxrockstoto";

        final ByteArrayOutputStream errBAOS = new ByteArrayOutputStream();
        final ByteArrayOutputStream outBAOS = new ByteArrayOutputStream();

        final ShellManagingExitCode shell = new ShellManagingExitCode(outBAOS, errBAOS);
        final SystemCall command = new SystemCall();
        command.setShell(shell);
        command.execute( new String[] { cmd });

        assertTrue(0 != shell.getExitStatus());
    }
}
