/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.pref;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.List;
import java.util.Properties;

import org.junit.Test;
import org.ow2.petals.cli.api.connection.ConnectionParameters;
import org.ow2.petals.cli.api.connection.exception.HostMissingException;
import org.ow2.petals.cli.api.connection.exception.InvalidPortException;
import org.ow2.petals.cli.api.pref.PreferenceFileException;
import org.ow2.petals.cli.api.pref.exception.DirectoryAsPreferenceFileException;
import org.ow2.petals.cli.api.pref.exception.IncorrectPropertyNameException;
import org.ow2.petals.cli.api.pref.exception.IncorrectPropertyValueException;
import org.ow2.petals.cli.api.pref.exception.MissingPropertyException;
import org.ow2.petals.cli.api.pref.exception.PreferenceFileNotFoundException;
import org.ow2.petals.cli.api.pref.exception.PreferenceFileNotReadableException;
import org.ow2.petals.cli.base.junit.AbstractConsoleInOutTest;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

public class AbstractPreferencesImplTest extends AbstractConsoleInOutTest {

    /**
     * The env variable containing the preference file
     */
    public final static String PREF_ENV_VAR = "PETALS_CLI_PREFS";

    /**
     * The default preferences file as resource
     */
    private final static String DEFAULT_PREF_RESOURCE = "default-preferences.properties";

    private class MyPreferenceImpl extends AbstractPreferencesImpl {

        protected MyPreferenceImpl(final String filePrefEnvVarName, final String filePrefAsDefaultResource) {
            super(filePrefEnvVarName, filePrefAsDefaultResource);
        }

        @Override
        protected ConnectionParameters createConnectionParameters(final Properties preferenceProperties,
                final String alias, final String host, final int port, final List<Exception> errors) {
            ConnectionParameters connectionParameters = null;
            try {
                if (host != null && port != -1) {
                    connectionParameters = new ConnectionParameters(host, port);
                }
            } catch (final UnknownHostException e) {
                errors.add(new IncorrectPropertyValueException(alias + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX,
                        host, e));
            } catch (final HostMissingException e) {
                errors.add(new IncorrectPropertyValueException(alias + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX,
                        host, e));
            } catch (final InvalidPortException e) {
                errors.add(new IncorrectPropertyValueException(alias + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX,
                        String.valueOf(port), e));
            }
            return connectionParameters;
        }
    }

    private void setUpUnreadablePrefFile() throws Exception {
        setUpUnreadablePrefFile(PREF_ENV_VAR);
    }

    private void setUpInvalidPrefFile() throws Exception {
        setUpInvalidPrefFile(PREF_ENV_VAR);
    }

    private void setUpDirectoryAsPrefFile() throws Exception {
        setUpDirectoryAsPrefFile(PREF_ENV_VAR);
    }

    private void setUpPrefFile(final Properties preferenceProperties) throws Exception {
        setUpPrefFile(PREF_ENV_VAR, preferenceProperties);
    }

    /**
     * <p>
     * Verify that the preferences file used when the environment variable is not set.
     * </p>
     * <p>
     * The expected preferences file is the default one, as resource.
     * </p>
     */
    @Test
    public final void testGetPreferenceFile_ByDefault() throws URISyntaxException, InvocationTargetException {

        final URL expectedPrefFileUrl = Thread.currentThread().getContextClassLoader()
                .getResource(DEFAULT_PREF_RESOURCE);
        final File expectedPrefFile = new File(expectedPrefFileUrl.toURI());

        assertNull("Env var " + PREF_ENV_VAR + "is set", System.getenv(PREF_ENV_VAR));
        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        final File preferenceFile = (File) ReflectionHelper.invokePrivateMethod(AbstractPreferencesImpl.class, pref,
                "getPreferenceFile", null, null);

        assertEquals("The preference file is not the expected one.", expectedPrefFile, preferenceFile);
    }

    /**
     * <p>
     * Verify that the preferences file used when the environment variable is set to an empty value.
     * </p>
     * <p>
     * The expected preferences file is the default one.
     * </p>
     */
    @Test
    public final void testGetPreferenceFile_WithEmptyEnvVar() throws Exception {

        envVars.set(PREF_ENV_VAR, "");

        final URL expectedPrefFileUrl = Thread.currentThread().getContextClassLoader()
                .getResource(DEFAULT_PREF_RESOURCE);
        final File expectedPrefFile = new File(expectedPrefFileUrl.toURI());

        assertNotNull("Env var " + PREF_ENV_VAR + " no set", System.getenv(PREF_ENV_VAR));
        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        final File preferenceFile = (File) ReflectionHelper.invokePrivateMethod(AbstractPreferencesImpl.class, pref,
                "getPreferenceFile", null, null);

        assertEquals("The preference file is not the expected one.", expectedPrefFile, preferenceFile);
    }

    /**
     * <p>
     * Verify the preferences file used when the environment variable is set to an unexisting file.
     * </p>
     * <p>
     * The expected preferences file is the one defined in the environment variable.
     * </p>
     */
    @Test
    public final void testGetPreferenceFile_WithInvalidEnvVar() throws Exception {

        setUpInvalidPrefFile();
        final String prefFile = System.getenv(PREF_ENV_VAR);
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);
        final File expectedPrefFile = new File(prefFile);

        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        final File preferenceFile = (File) ReflectionHelper.invokePrivateMethod(AbstractPreferencesImpl.class, pref,
                "getPreferenceFile", null, null);

        assertEquals("The preference file is not the expected one.", expectedPrefFile, preferenceFile);
    }

    /**
     * <p>
     * Verify preferences file used when the environment variable is set to an existing file.
     * </p>
     * <p>
     * The expected preferences file is the one defined in the environment variable.
     * </p>
     */
    @Test
    public final void testGetPreferenceFile_WithSetEnvVar() throws Exception {

        setUpPrefFile(new Properties());
        final String prefFile = System.getenv(PREF_ENV_VAR);

        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);
        final File expectedPrefFile = new File(prefFile);

        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        final File preferenceFile = (File) ReflectionHelper.invokePrivateMethod(AbstractPreferencesImpl.class, pref,
                "getPreferenceFile", null, null);

        assertEquals("The preference file is not the expected one.", expectedPrefFile, preferenceFile);
    }

    /**
     * <p>
     * Verify the read of preferences file when the preference file does not exist.
     * </p>
     * <p>
     * Expected result: the error {@link PreferenceFileNotFoundException} occurs.
     * </p>
     */
    @Test(expected = PreferenceFileNotFoundException.class)
    public final void testInitializePreferenceParameters_WithUnexistingPreferenceFile() throws Exception {

        setUpInvalidPrefFile();
        final String prefFile = System.getenv(PREF_ENV_VAR);
        
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);
        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        pref.intializePreferenceParameters();
    }

    /**
     * <p>
     * Verify the read of preferences file when the preference file is not readable because of ACL.
     * </p>
     * <p>
     * Expected result: the error {@link PreferenceFileNotReadableException} occurs.
     * </p>
     */
    @Test(expected = PreferenceFileNotReadableException.class)
    public final void testInitializePreferenceParameters_WithNotReadablePreferenceFile() throws Exception {

        setUpUnreadablePrefFile();
        final String prefFile = System.getenv(PREF_ENV_VAR);

        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);
        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        pref.intializePreferenceParameters();
    }

    /**
     * <p>
     * Verify the read of preferences file when a directory is given as preference file.
     * </p>
     * <p>
     * Expected result: the error {@link DirectoryAsPreferenceFileException} occurs.
     * </p>
     */
    @Test(expected = DirectoryAsPreferenceFileException.class)
    public final void testInitializePreferenceParameters_WithDirectoryAsPreferenceFile() throws Exception {

        setUpDirectoryAsPrefFile();
        final String directoryAsPrefFile = System.getenv(PREF_ENV_VAR);

        assertNotNull("Env var " + PREF_ENV_VAR + " no set", directoryAsPrefFile);
        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        pref.intializePreferenceParameters();
    }

    /**
     * <p>
     * Verify the default connection parameters when a default alias is correctly set with connection parameters.
     * </p>
     * <p>
     * Expected results: the default connection parameters are the ones correctly read from the default alias
     * </p>
     */
    @Test
    public final void testInitializeDefaultConnectionAliasWithoutErrors() throws Exception {

        
        final String alias1 = "server1";
        final String alias2 = "server2";
        final Properties preferenceProperties = new Properties();

        final String defaultHost = "localhost";
        final int defaultPort = 7458;
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX, defaultHost);
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX,
                String.valueOf(defaultPort));
        preferenceProperties.setProperty(alias2 + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX, "127.0.0.2");
        preferenceProperties.setProperty(alias2 + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX, "1234");
        preferenceProperties.setProperty(AbstractPreferencesImpl.DEFAULT_ALIAS_PROPERTY, alias1);


        setUpPrefFile(preferenceProperties);
        final String prefFile = System.getenv(PREF_ENV_VAR);
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);

        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        pref.intializePreferenceParameters();
        final ConnectionParameters connectionParameters = pref.getDefaultConnectionParameters();
        assertNotNull(connectionParameters);
        assertEquals(defaultHost, connectionParameters.getHost());
        assertEquals(defaultPort, connectionParameters.getPort());
    }

    /**
     * <p>
     * Verify the default connection parameters when a no default alias is set.
     * </p>
     * <p>
     * Expected results: an error occurs because the default alias property is missing
     * </p>
     */
    @Test(expected = PreferenceFileException.class)
    public final void testInitializeDefaultConnectionAliasWithNoDefaultConnectionAlias() throws Exception {

        final Properties preferenceProperties = new Properties();

        final String alias1 = "server1";
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX, "test1");
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX, "7458");

        setUpPrefFile(preferenceProperties);
        final String prefFile = System.getenv(PREF_ENV_VAR);
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);

        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        try {
            pref.intializePreferenceParameters();
        } catch (final PreferenceFileException e) {
            assertTrue(e.getMessage().contains(
                    String.format(MissingPropertyException.MESSAGE_PATTERN,
                            AbstractPreferencesImpl.DEFAULT_ALIAS_PROPERTY)));
            throw e;
        }
    }

    /**
     * <p>
     * Verify the default connection parameters when preferences are empty.
     * </p>
     * <p>
     * Expected results: an error occurs because the default alias property is missing
     * </p>
     */
    @Test(expected = PreferenceFileException.class)
    public final void testInitializeDefaultConnectionAliasWithEmptyPreferenceFile() throws Exception {

        final Properties preferenceProperties = new Properties();

        setUpPrefFile(preferenceProperties);
        final String prefFile = System.getenv(PREF_ENV_VAR);
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);

        final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
        try {
            pref.intializePreferenceParameters();
        } catch (final PreferenceFileException e) {
            assertTrue(e.getMessage().contains(
                    String.format(MissingPropertyException.MESSAGE_PATTERN,
                            AbstractPreferencesImpl.DEFAULT_ALIAS_PROPERTY)));
            throw e;
        }
    }

    /**
     * <p>
     * Verify the connection parameters with errors:
     * <ul>
     * <li>a port is not a number</li>
     * <li>a suffix is missing on an alias</li>
     * <li>a suffix is unknown</li>
     * </ul>
     * </p>
     * <p>
     * Expected results: errors are correctly reported
     * </p>
     */
    @Test(expected = PreferenceFileException.class)
    public final void testInitializePreferenceConnectionParametersWithErrors() throws Exception {

        final Properties preferenceProperties = new Properties();

        final String alias1 = "server1";
        final String invalidPort = "745ds";
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.HOST_PROPERTY_SUFFIX, "localhost");
        preferenceProperties.setProperty(alias1 + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX, invalidPort);

        final String aliasfd = "servefd";
        preferenceProperties.setProperty(aliasfd, "testIncorrect");

        final String alias5 = "server5";
        final String unknownSuffix = ".usernme";
        preferenceProperties.setProperty(alias5 + ".usernme", "userIncorect");

        setUpPrefFile(preferenceProperties);
        final String prefFile = System.getenv(PREF_ENV_VAR);
        assertNotNull("Env var " + PREF_ENV_VAR + " no set", prefFile);

        try {
            final AbstractPreferencesImpl pref = new MyPreferenceImpl(PREF_ENV_VAR, DEFAULT_PREF_RESOURCE);
            pref.intializePreferenceParameters();
        } catch (final PreferenceFileException e) {
            assertTrue(e.getMessage().contains(
                    String.format(IncorrectPropertyValueException.MESSAGE_PATTERN, invalidPort, alias1
                            + AbstractPreferencesImpl.PORT_PROPERTY_SUFFIX)));
            assertTrue(e.getMessage().contains(String.format(IncorrectPropertyNameException.MESSAGE_PATTERN, aliasfd)));
            assertTrue(e.getMessage().contains(
                    String.format(IncorrectPropertyNameException.MESSAGE_PATTERN, alias5 + unknownSuffix)));
            throw e;
        }
    }
}
