/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.shell;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.Map;

import org.apache.commons.cli.CommandLine;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.EnvironmentVariables;
import org.ow2.petals.cli.api.command.AbstractCommand;
import org.ow2.petals.cli.api.command.Command;
import org.ow2.petals.cli.api.command.ConnectionCommand;
import org.ow2.petals.cli.api.command.exception.CommandException;
import org.ow2.petals.cli.api.command.exception.ConnectionErrorException;
import org.ow2.petals.cli.api.connection.AuthenticatedConnectionParameters;
import org.ow2.petals.cli.api.connection.ConnectionParameters;
import org.ow2.petals.cli.api.connection.exception.InvalidConnectionParameterException;
import org.ow2.petals.cli.api.pref.PreferenceFileException;
import org.ow2.petals.cli.api.shell.exception.ConnectionCommandAlreadyRegisteredException;
import org.ow2.petals.cli.api.shell.exception.DuplicatedCommandException;
import org.ow2.petals.cli.api.shell.exception.ShellException;

/**
 * Unit tests of {@link AbstractShell}
 * 
 * @author Sebastien Andre - EBM WebSourcing
 * @author Christophe DENEUX - EBM WebSourcing
 */
public class AbstractShellTest {


    final private static String INTERPOLATE_TEST_ENV_VAR = "INTERPOLATE_TEST_ENV_VAR";

    final private static String INTERPOLATE_TEST_ENV_VAR_VALUE = "interpolate-test-env-var-value";

    @Rule
    public final EnvironmentVariables envVars = new EnvironmentVariables();

    final private static String INTERPOLATE_TEST_PROP_NAME = "var";

    final private static String INTERPOLATE_TEST_PROP_VALUE = "value";

    private TestShell shell;

    private class TestShell extends AbstractShell {

        public TestShell() throws ShellException {
            super(System.out, System.err, false, false, null, null, null);
        }

        @Override
        public boolean isComment(String line) {
            return super.isComment(line);
        }

        @Override
        public void run() {
            // NOP
        }

        @Override
        public String askQuestion(String question, boolean isReplyPassword) throws ShellException {
            return null;
        }

        @Override
        public boolean confirms(String message) throws ShellException {
            return false;
        }
    }

    private class TestCommand extends AbstractCommand {

        @Override
        public void doExecute(final String[] args) throws CommandException {
            // NOP
        }
    }

    private class TestConnectionCommand extends AbstractCommand implements ConnectionCommand {

        public TestConnectionCommand() {
            super();
        }

        public TestConnectionCommand(final String name) {
            super(name);
        }

        @Override
        public void doExecute(final String[] args) throws CommandException {
            // NOP
        }

        @Override
        public AuthenticatedConnectionParameters connect(final ConnectionParameters connectionParameters,
                final boolean isYesFlagEnabledAtCmdLevel) throws ConnectionErrorException, CommandException {
            return null;
        }

        @Override
        public ConnectionParameters parseConnectionParameters(final CommandLine cmd, final boolean useDefaultConnection)
                throws CommandException, PreferenceFileException, InvalidConnectionParameterException {
            return null;
        }
    }

    @Before
    public void setUp() throws Exception {
        this.shell = new TestShell();
        this.envVars.set(INTERPOLATE_TEST_ENV_VAR, INTERPOLATE_TEST_ENV_VAR_VALUE);
    }

    @After
    public void tearDown() throws Exception {
    }

    @Test
    public final void testIsComment() {
        assertTrue(shell.isComment("#   "));
        assertTrue(shell.isComment(""));
        assertTrue(shell.isComment("   "));
        assertTrue(shell.isComment("           "));
        assertTrue(shell.isComment("    #"));
        assertTrue(shell.isComment("    ## ### ## "));
        assertFalse(shell.isComment("joe"));
        assertFalse(shell.isComment("  joe# "));
        assertFalse(shell.isComment(".#"));
    }

    @Test
    public final void testInterpolate() {

        assertEquals("booo", this.shell.interpolate("booo"));
        assertEquals("   space   ", this.shell.interpolate("   space   "));

        System.clearProperty(INTERPOLATE_TEST_PROP_NAME);
        System.clearProperty("var2");
        assertEquals("${" + INTERPOLATE_TEST_PROP_NAME + "}",
                this.shell.interpolate("${" + INTERPOLATE_TEST_PROP_NAME + "}"));
        assertEquals("${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME + "}",
                this.shell.interpolate("${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME + "}"));
        assertEquals(
                "${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME + "} ${var2}",
                this.shell.interpolate("${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME
                        + "} ${var2}"));

        System.setProperty(INTERPOLATE_TEST_PROP_NAME, INTERPOLATE_TEST_PROP_VALUE);
        assertEquals(INTERPOLATE_TEST_PROP_VALUE, this.shell.interpolate("${" + INTERPOLATE_TEST_PROP_NAME + "}"));
        assertEquals(" " + INTERPOLATE_TEST_PROP_VALUE + " ",
                this.shell.interpolate(" ${" + INTERPOLATE_TEST_PROP_NAME + "} "));
        assertEquals(" " + INTERPOLATE_TEST_PROP_VALUE + " " + INTERPOLATE_TEST_PROP_VALUE + " ",
                this.shell.interpolate(" ${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME + "} "));
        assertEquals(
                " " + INTERPOLATE_TEST_PROP_VALUE + " " + INTERPOLATE_TEST_PROP_VALUE + " ${var2} ",
                this.shell.interpolate(" ${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME
                        + "} ${var2} "));

        assertEquals("${" + INTERPOLATE_TEST_PROP_NAME + "}",
                this.shell.interpolate("\\${" + INTERPOLATE_TEST_PROP_NAME + "}"));
        assertEquals(
                " ${" + INTERPOLATE_TEST_PROP_NAME + "} " + INTERPOLATE_TEST_PROP_VALUE + " \\${"
                        + INTERPOLATE_TEST_PROP_NAME + "} ",
                this.shell.interpolate(" \\${" + INTERPOLATE_TEST_PROP_NAME + "} ${" + INTERPOLATE_TEST_PROP_NAME
                        + "} \\\\${" + INTERPOLATE_TEST_PROP_NAME + "} "));

        // Environment variables
        assertNotNull("Env var INTERPOLATE_TEST_ENV_VAR no set", System.getenv("INTERPOLATE_TEST_ENV_VAR"));
        assertEquals(INTERPOLATE_TEST_ENV_VAR_VALUE, this.shell.interpolate("${env:INTERPOLATE_TEST_ENV_VAR}"));
        assertEquals(" " + INTERPOLATE_TEST_ENV_VAR_VALUE + " ",
                this.shell.interpolate(" ${env:INTERPOLATE_TEST_ENV_VAR} "));
        assertEquals(" " + INTERPOLATE_TEST_ENV_VAR_VALUE + " " + INTERPOLATE_TEST_ENV_VAR_VALUE + " ",
                this.shell.interpolate(" ${env:INTERPOLATE_TEST_ENV_VAR} ${env:INTERPOLATE_TEST_ENV_VAR} "));
        assertEquals(" " + INTERPOLATE_TEST_ENV_VAR_VALUE + " " + INTERPOLATE_TEST_ENV_VAR_VALUE + " ${var2} ",
                this.shell.interpolate(" ${env:INTERPOLATE_TEST_ENV_VAR} ${env:INTERPOLATE_TEST_ENV_VAR} ${var2} "));

        assertEquals("${env:INTERPOLATE_TEST_ENV_VAR}", this.shell.interpolate("\\${env:INTERPOLATE_TEST_ENV_VAR}"));
        assertEquals(
                " ${env:INTERPOLATE_TEST_ENV_VAR} " + INTERPOLATE_TEST_ENV_VAR_VALUE
                        + " \\${env:INTERPOLATE_TEST_ENV_VAR} ",
                this.shell
                        .interpolate(" \\${env:INTERPOLATE_TEST_ENV_VAR} ${env:INTERPOLATE_TEST_ENV_VAR} \\\\${env:INTERPOLATE_TEST_ENV_VAR} "));

        // Environment variables and properties
        assertEquals(" " + INTERPOLATE_TEST_PROP_VALUE + " " + INTERPOLATE_TEST_ENV_VAR_VALUE + " ",
                this.shell.interpolate(" ${" + INTERPOLATE_TEST_PROP_NAME + "} ${env:INTERPOLATE_TEST_ENV_VAR} "));
    }

    /**
     * Check the register method against a command.
     */
    @Test
    public final void testRegister_Command() throws DuplicatedCommandException {

        final Command myCommand = new TestCommand();
        this.shell.registersCommand(myCommand);

        final Map<String, Command> commands = this.shell.getCommands();
        assertNotNull(commands);
        assertEquals(1, commands.size());

        final Command registeredCommand = commands.get(myCommand.getName());
        assertNotNull(registeredCommand);

        assertNull(this.shell.getConnectionCommand());
    }

    /**
     * Check the register method against a duplicated command.
     */
    @Test(expected = DuplicatedCommandException.class)
    public final void testRegister_DuplicatedCommand() throws DuplicatedCommandException {

        final Command myCommand = new TestCommand();
        this.shell.registersCommand(myCommand);

        final Map<String, Command> commands = this.shell.getCommands();
        assertNotNull(commands);
        assertEquals(1, commands.size());

        try {
            this.shell.registersCommand(myCommand);
        } catch (final DuplicatedCommandException e) {
            assertEquals(e.getCommandName(), myCommand.getName());
            throw e;
        }
    }

    /**
     * Check the registration of a valid connection command.
     */
    @Test
    public final void testRegister_ConnectionCommand() throws DuplicatedCommandException,
            ConnectionCommandAlreadyRegisteredException, IllegalArgumentException {

        final ConnectionCommand myConnectCommand = new TestConnectionCommand();
        this.shell.registersConnectionCommand(myConnectCommand);

        final Map<String, Command> commands = this.shell.getCommands();
        assertNotNull(commands);
        assertEquals(1, commands.size());

        final Command registeredCommand = commands.get(myConnectCommand.getName());
        assertNotNull(registeredCommand);

        assertNotNull(this.shell.getConnectionCommand());
    }

    /**
     * Check the registration of a valid connection command when another command was previously registered with the same
     * name
     */
    @Test(expected = DuplicatedCommandException.class)
    public final void testRegister_DuplicatedConnectionCommand() throws DuplicatedCommandException,
            ConnectionCommandAlreadyRegisteredException, IllegalArgumentException {

        final Command myCommand = new TestCommand();
        this.shell.registersCommand(myCommand);

        final Map<String, Command> commands = this.shell.getCommands();
        assertNotNull(commands);
        assertEquals(1, commands.size());

        final ConnectionCommand myConnectionCommand = new TestConnectionCommand(myCommand.getName());
        try {
            this.shell.registersConnectionCommand(myConnectionCommand);
        } catch (final DuplicatedCommandException e) {
            assertEquals(e.getCommandName(), myCommand.getName());
            throw e;
        }
    }

    /**
     * Check the registration of a second valid connection command
     */
    @Test(expected = ConnectionCommandAlreadyRegisteredException.class)
    public final void testRegister_SecondConnectionCommand() throws DuplicatedCommandException,
            ConnectionCommandAlreadyRegisteredException, IllegalArgumentException {

        final ConnectionCommand myConnectionCommand = new TestConnectionCommand();
        this.shell.registersConnectionCommand(myConnectionCommand);

        final ConnectionCommand anotherConnectionCommand = new TestConnectionCommand("another-connection-command");
        try {
            this.shell.registersConnectionCommand(anotherConnectionCommand);
        } catch (final ConnectionCommandAlreadyRegisteredException e) {
            assertEquals(e.getCommandName(), anotherConnectionCommand.getName());
            throw e;
        }
    }

}
