/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.shell;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.net.UnknownHostException;

import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.SystemErrRule;
import org.ow2.petals.cli.api.command.AbstractCommand;
import org.ow2.petals.cli.api.connection.AuthenticatedConnectionParameters;
import org.ow2.petals.cli.api.connection.ConnectionParameters;
import org.ow2.petals.cli.api.connection.exception.InvalidConnectionParameterException;
import org.ow2.petals.cli.api.shell.exception.ShellException;
import org.ow2.petals.cli.shell.command.CommandExecutedWithError;
import org.ow2.petals.cli.shell.command.CommandNoArg;
import org.ow2.petals.cli.shell.command.CommandResult;

/**
 * Unit tests of {@link PetalsCli}
 * 
 * @author Christophe DENEUX - Linagora
 */
public class PetalsCliTest {

    /**
     * The shell {@link PetalsCli} establishes automatically a connection. Here,
     * the default connection parameters.
     */
    private final ConnectionParameters defaultConnectionParameters;

    @Rule
    public final SystemErrRule systemErr = new SystemErrRule().muteForSuccessfulTests().enableLog();

    public PetalsCliTest() {
        try {
            this.defaultConnectionParameters = new AuthenticatedConnectionParameters("localhost", 7700) {
            };
        } catch (final UnknownHostException e) {
            // NOP, the exception can not occur
            throw new RuntimeException("An unexpected exception occurs", e);
        } catch (final InvalidConnectionParameterException e) {
            // NOP, the exception can not occur
            throw new RuntimeException("An unexpected exception occurs", e);
        }
    }

    /**
     * Check the run method against an not-registered entered command.
     */
    @Test
    public final void testRun_WithUnregisteredCommand() {

        final String unknownCommand = "unknowncommand";
        final AbstractShell shell = new PetalsCli(new String[] { unknownCommand }, false, false, null, null, null);

        shell.setConnectionParameters(this.defaultConnectionParameters);
        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.startsWith("ERROR:"));
        assertTrue(error.contains(unknownCommand));
        assertEquals("Invalid exit code", 1, shell.getExitStatus());


    }

    /**
     * Check the run method against an right entered command.
     * 
     * @throws ShellException
     */
    @Test
    public final void testRun_RegisteredCommand() throws ShellException {

        final AbstractShell shell = new PetalsCli(new String[] { CommandNoArg.class.getSimpleName().toLowerCase() },
                false, false, null, null, null);

        final CommandResult cmdRes = new CommandResult();

        shell.registersCommand(new CommandNoArg(cmdRes));

        shell.setConnectionParameters(this.defaultConnectionParameters);
        shell.run();

        final String error = systemErr.getLog();
        assertTrue(error.isEmpty());
        assertTrue(cmdRes.isExecuted());
        assertEquals("Invalid exit code", 0, shell.getExitStatus());

    }

    /**
     * Check the run method against an command entered with error.
     * 
     * @throws ShellException
     */
    @Test
    public final void testRun_CommandEnteredWithError() throws ShellException {


        final CommandResult cmdRes = new CommandResult();
        final AbstractCommand command = new CommandNoArg(cmdRes);

        final AbstractShell shell = new PetalsCli(new String[] { command.getName(), "arg1" }, false, false, null, null,
                null);

        shell.registersCommand(command);

        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertTrue("The command usage is missing in error message", error.contains(command.getUsage()));
        assertEquals("Invalid exit code", 1, shell.getExitStatus());


    }

    /**
     * Check the run method against an command executed with error.
     * 
     * @throws ShellException
     */
    @Test
    public final void testRun_CommandExecutedWithError() throws ShellException {

        final CommandResult cmdRes = new CommandResult();
        final AbstractCommand command = new CommandExecutedWithError(cmdRes);

        final AbstractShell shell = new PetalsCli(
                new String[] { command.getName(), "-" + CommandExecutedWithError.DUMMY_SHORT_OPTION }, false, false,
                null, null, null);

        shell.registersCommand(command);

        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertTrue("The command usage is contained in error message", !error.contains(command.getUsage()));
        assertEquals("Invalid exit code", 2, shell.getExitStatus());


    }
}
