/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.shell;

import static org.hamcrest.Matchers.allOf;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.startsWith;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.IOException;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.ow2.petals.cli.api.command.AbstractCommand;
import org.ow2.petals.cli.api.shell.exception.DuplicatedCommandException;
import org.ow2.petals.cli.api.shell.exception.ShellException;
import org.ow2.petals.cli.base.junit.AbstractConsoleInOutTest;
import org.ow2.petals.cli.base.junit.shell.PetalsInteractiveCliUtils;
import org.ow2.petals.cli.command.Exit;
import org.ow2.petals.cli.shell.command.CommandExecutedWithError;
import org.ow2.petals.cli.shell.command.CommandNoArg;
import org.ow2.petals.cli.shell.command.CommandResult;

/**
 * Unit tests of {@link PetalsInteractiveCli}
 *
 * @author Christophe DENEUX - EBM Websourcing
 */
public class PetalsInteractiveCliTest extends AbstractConsoleInOutTest {

    private void runCli(final AbstractShell shell) {
        this.runCli(new Runnable() {
            @Override
            public void run() {
                try {
                    shell.registersCommand(new Exit());
                } catch (DuplicatedCommandException | IllegalArgumentException e) {
                    throw new AssertionError(e);
                }
                shell.run();
            }
        });
    }

    private void waitThreadOrFail(final AbstractShell shell) throws IOException {
        systemIn.writeLine("exit");
        super.waitThreadOrFail();
    }

    @BeforeClass
    public static void beforeClass() {
        System.setProperty("jline.shutdownhook", "true");
        PetalsInteractiveCliUtils.hackForWindows(true);
    }

    @AfterClass
    public static void afterClass() {
        System.clearProperty("jline.shutdownhook");
        PetalsInteractiveCliUtils.hackForWindows(false);
    }

    /**
     * Checks the run method against an unregistered entered command.
     */
    @Test
    public final void testRun_WithUnregisteredCommand() throws IOException {

        final String unknownCommand = "unknowncommand";

        final AbstractShell shell = new PetalsInteractiveCli(false, false, null, null, null);

        runCli(shell);

        systemIn.writeLine(unknownCommand);

        waitThreadOrFail(shell);

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.startsWith("ERROR:"));
        assertTrue(error.contains(unknownCommand));
        assertEquals("Invalid exit code", 0, shell.getExitStatus());


    }

    /**
     * Checks the run method against a right entered command.
     */
    @Test
    public final void testRun_RegisteredCommand() throws IOException, DuplicatedCommandException {

        final AbstractShell shell = new PetalsInteractiveCli(false, false, null, null, null);

        final CommandResult cmdRes = new CommandResult();

        final CommandNoArg command = new CommandNoArg(cmdRes);

        shell.registersCommand(command);

        runCli(shell);

        systemIn.writeLine(command.getName());

        waitThreadOrFail(shell);

        final String error = systemErr.getLog();
        assertTrue(error.isEmpty());
        assertTrue(cmdRes.isExecuted());
        assertEquals("Invalid exit code", 0, shell.getExitStatus());


    }

    /**
     * Checks the run method against a command entered with error.
     */
    @Test
    public final void testRun_CommandWithError() throws IOException, DuplicatedCommandException {

        final CommandResult cmdRes = new CommandResult();
        final AbstractCommand command = new CommandNoArg(cmdRes);

        final AbstractShell shell = new PetalsInteractiveCli(false, false, null, null, null);

        shell.registersCommand(command);

        runCli(shell);

        systemIn.writeLine(command.getName() + " arg1");

        checkErr(
                allOf(startsWith("ERROR on command '" + command.getName() + "':"), containsString(command.getUsage())));

        waitThreadOrFail(shell);

        assertEquals("Invalid exit code", 0, shell.getExitStatus());

    }

    /**
     * Checks the run method against a command executed with error.
     * @throws ShellException
     */
    @Test
    public final void testRun_CommandExecutedWithError() throws IOException, ShellException {

        final CommandResult cmdRes = new CommandResult();
        final AbstractCommand command = new CommandExecutedWithError(cmdRes);

        final AbstractShell shell = new PetalsInteractiveCli(false, false, null, null, null);

        shell.registersCommand(command);

        runCli(shell);

        systemIn.writeLine(command.getName() + " -" + CommandExecutedWithError.DUMMY_SHORT_OPTION);

        waitThreadOrFail(shell);

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue("The command is missing in error message",
                error.startsWith("ERROR on command '" + command.getName() + "':"));
        assertTrue("The command usage is contained in error message", !error.contains(command.getUsage()));
        assertEquals("Invalid exit code", 0, shell.getExitStatus());

    }
}
