/**
 * Copyright (c) 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.cli.shell;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.UnknownHostException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.MissingArgumentException;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.UnrecognizedOptionException;
import org.junit.Test;
import org.ow2.petals.cli.AbstractMain;
import org.ow2.petals.cli.api.connection.AuthenticatedConnectionParameters;
import org.ow2.petals.cli.api.connection.ConnectionParameters;
import org.ow2.petals.cli.api.connection.exception.InvalidConnectionParameterException;
import org.ow2.petals.cli.api.shell.exception.ShellException;
import org.ow2.petals.cli.base.junit.AbstractConsoleInOutTest;
import org.ow2.petals.cli.command.Help;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit test of {@link ShellFactory}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ShellFactoryTest extends AbstractConsoleInOutTest {

    /**
     * The shell {@link PetalsCli} establishes automatically a connection. Here,
     * the default connection parameters.
     */
    private final ConnectionParameters defaultConnectionParameters;

    public ShellFactoryTest() {
        try {
            this.defaultConnectionParameters = new AuthenticatedConnectionParameters("localhost", 7700) {
            };
        } catch (final UnknownHostException e) {
            // NOP, the exception can not occur
            throw new RuntimeException("An unexpected exception occurs", e);
        } catch (InvalidConnectionParameterException e) {
            // NOP, the exception can not occur
            throw new RuntimeException("An unexpected exception occurs", e);
        }
    }

    private AbstractShell getShell(final String[] commandLineArgs) throws ParseException, FileNotFoundException,
            ShellException, InvocationTargetException {

        final CommandLineParser parser = new DefaultParser();
        final AbstractMain main = new AbstractMain(null) {

            @Override
            protected String getUsageHeader() {
                return null;
            }

            @Override
            protected String getPromptBase() {
                return "a-base-prompt";
            }
        };
        final Options options = (Options) ReflectionHelper.invokePrivateMethod(AbstractMain.class, main,
                "createOptions", null, null);
        final CommandLine cmd = parser.parse(options, commandLineArgs);
        return ShellFactory.getInstance().newShell(cmd, commandLineArgs, null, null, "abtract-shell");

    }

    /**
     * Check that Petals CLI is launched in interactive mode using following
     * command line: <code>petals-cli -C</code>
     */
    @Test
    public void testnewShell_interactiveMode() throws ParseException, FileNotFoundException, ShellException,
            InvocationTargetException {

        final AbstractShell shell = this.getShell(new String[] { "-" + ShellFactory.CONSOLE_SHORT_OPTION });
        assertTrue("Interactive mode not detected", shell instanceof PetalsInteractiveCli);

    }

    /**
     * Check the error management when Petals CLI is launched in interactive
     * mode with unknown argument before, using following command line:
     * <code>petals-cli --unknown-arg -C</code>
     */
    @Test
    public void testnewShell_interactiveMode_error_00() throws Exception {
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { unrecognizedArg, "--" + ShellFactory.CONSOLE_SHORT_OPTION });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e
                    .getMessage().contains(unrecognizedArg));
        }
    }

    /**
     * Check the error management when Petals CLI is launched in interactive
     * mode with unknown argument after, using following command line:
     * <code>petals-cli -C --unknown-arg</code>
     */
    @Test
    public void testnewShell_interactiveMode_error_01() throws Exception {
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { "-" + ShellFactory.CONSOLE_SHORT_OPTION, unrecognizedArg });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e
                    .getMessage().contains(unrecognizedArg));
        }
    }

    /**
     * Check that Petals CLI is launched in command mode using following command
     * line: <code>petals-cli -c -- help</code>
     */
    @Test
    public void testnewShell_commandMode() throws ParseException, FileNotFoundException, ShellException,
            InvocationTargetException {

        final AbstractShell shell = this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION,
                "--",
                "help" });
        assertTrue("command mode not detected", shell instanceof PetalsCli);
    }

    /**
     * Check the error management when Petals CLI is launched in command mode
     * with unknown argument before the command flag, using following command
     * line: <code>petals-cli --unknown-arg -c -- help</code>
     */
    @Test
    public void testnewShell_commandMode_error_00() throws Exception {
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { unrecognizedArg, "-" + ShellFactory.COMMAND_SHORT_OPTION,
                    "--",
                    "help" });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e
                    .getMessage().contains(unrecognizedArg));
        }
    }

    /**
     * Check the error management when Petals CLI is launched in command mode
     * with unknown argument between the operand flag and the command, using
     * following command line: <code>petals-cli -c -- --unknown-arg help</code>
     */
    @Test
    public void testnewShell_commandMode_error_01() throws FileNotFoundException, ParseException, ShellException,
            InvocationTargetException {

        final String unrecognizedArg = "--unknown-arg";
        final AbstractShell shell = this
                .getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--", unrecognizedArg, "help" });

        assertTrue("command mode not detected", shell instanceof PetalsCli);

        shell.setConnectionParameters(this.defaultConnectionParameters);
        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.contains(unrecognizedArg));
        assertTrue(error.contains("not registered"));


    }

    /**
     * Check the error management when Petals CLI is launched in command mode
     * with unknown argument between the command flag and the operand flag,
     * using following command line:
     * <code>petals-cli -c --unknown-arg -- help</code>
     */
    @Test
    public void testnewShell_commandMode_error_02() throws FileNotFoundException, ParseException, ShellException,
            InvocationTargetException {
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, unrecognizedArg,
                    "--", "help" });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e.getMessage()
                    .contains(unrecognizedArg));
        }

    }

    /**
     * Check the error management when Petals CLI is launched in command mode
     * with unknown argument after the command, using following command line:
     * <code>petals-cli -c -- help --unknown-arg</code>
     */
    @Test
    public void testnewShell_commandMode_error_03() throws Exception {

        final String unrecognizedArg = "--unknown-arg";

        final AbstractShell shell = this.getShell(new String[] {
                "-" + ShellFactory.COMMAND_SHORT_OPTION,
                "--",
                new Help().getName(),
                unrecognizedArg });

        assertTrue("command mode not detected", shell instanceof PetalsCli);

        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.contains(unrecognizedArg));
        assertTrue(error.contains("unknown command"));

    }

    /**
     * <p>
     * Check the error management when Petals CLI is launched in command mode
     * without operand flag and without command, using following command line:
     * <code>petals-cli -c</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error about the missing command.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testnewShell_commandMode_error_04() throws Exception {

        try {
            this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION });
            fail("ParseException is not thrown");
        } catch (final ParseException e) {
            // Expected exception
            assertTrue("It's not the expected error message.", e.getMessage().contains(ShellFactory.MISSING_COMMAND));
        }
    }

    /**
     * <p>
     * Check the error management when Petals CLI is launched in command mode
     * with operand flag but without command, using following command line:
     * <code>petals-cli -c --</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error about the missing command.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testnewShell_commandMode_error_05() throws Exception {

        try {
            this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "--" });
            fail("ParseException is not thrown");
        } catch (final ParseException e) {
            // Expected exception
            assertTrue("It's not the expected error message.", e.getMessage().contains(ShellFactory.MISSING_COMMAND));
        }
    }

    /**
     * <p>
     * Check the error management when Petals CLI is launched in command mode
     * without operand flag but with command needing no arguments, using
     * following command line: <code>petals-cli -c help</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error thrown.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testnewShell_commandMode_error_06() throws Exception {

        final AbstractShell shell = this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION,
                "help" });

        assertTrue("command mode not detected", shell instanceof PetalsCli);
    }

    /**
     * <p>
     * Check the error management when Petals CLI is launched in command mode
     * without operand flag but with command needing arguments, using following
     * command line: <code>petals-cli -c deploy -D myprop=myvalue</code>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>error about the unrecognized argument by Petals CLI, its an argument
     * of the command.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testnewShell_commandMode_error_07() throws Exception {

        try {
            this.getShell(new String[] { "-" + ShellFactory.COMMAND_SHORT_OPTION, "deplopy", "-D",
                    "myprop=myvalue" });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e.getMessage()
                    .contains("-D"));
        }
    }

    /**
     * Check that Petals CLI is launched in file mode using following command
     * line: <code>petals-cli --file existing-file</code>
     */
    @Test
    public void testnewShell_fileMode() throws ParseException, ShellException, IOException, InvocationTargetException {
        final File tmpFile = File.createTempFile("petals-cli", "commands");

        try {
            final AbstractShell shell = this.getShell(new String[] { "--" + ShellFactory.FILE_LONG_OPTION,
                    tmpFile.getAbsolutePath() });
            assertTrue("file mode not detected", shell instanceof PetalsFileScriptShell);
        } finally {
            tmpFile.delete();
        }
    }

    /**
     * Check the error management when Petals CLI is launched in file mode
     * without filename, using following command line:
     * <code>petals-cli --file</code>
     */
    @Test
    public void testnewShell_fileMode_error_00() throws ParseException, FileNotFoundException, ShellException,
            InvocationTargetException {

        try {
            final AbstractShell shell = this
                    .getShell(new String[] { "--" + ShellFactory.FILE_LONG_OPTION });
            assertTrue("file mode not detected", shell instanceof PetalsFileScriptShell);
            fail("MissingArgumentException is not thrown");
        } catch (final MissingArgumentException e) {
            assertEquals("Wrong option in error", ShellFactory.FILE_LONG_OPTION, e.getOption()
                    .getLongOpt());
        }
    }

    /**
     * Check the error management when Petals CLI is launched in file mode with
     * an unexisting filename, using following command line:
     * <code>petals-cli --file unexisting-filename</code>
     */
    @Test
    public void testnewShell_fileMode_error_01() throws Exception {

        final String unexistingFileName = "unexisting-file";
        final AbstractShell shell = this
                .getShell(new String[] { "--" + ShellFactory.FILE_LONG_OPTION, unexistingFileName });

        assertTrue("file mode not detected", shell instanceof PetalsFileScriptShell);

        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.contains("FileNotFoundException"));
        assertTrue(error.contains(unexistingFileName));


    }

    /**
     * Check the error management when Petals CLI is launched in file mode with
     * unknown argument before, using following command line:
     * <code>petals-cli --unknown-arg --file existing-filename</code>
     */
    @Test
    public void testnewShell_fileMode_error_02() throws Exception {
        final File tmpFile = File.createTempFile("petals-cli", "commands");
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { unrecognizedArg, "--" + ShellFactory.FILE_LONG_OPTION,
                    tmpFile.getAbsolutePath() });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e
                    .getMessage().contains(unrecognizedArg));
        }
    }

    /**
     * Check the error management when Petals CLI is launched in file mode with
     * unknown argument between, using following command line:
     * <code>petals-cli --file --unknown-arg existing-filename</code>
     */
    @Test
    public void testnewShell_fileMode_error_03() throws Exception {

        final File tmpFile = File.createTempFile("petals-cli", "commands");
        final String unrecognizedArg = "--unknown-arg";
        final AbstractShell shell = this.getShell(
                new String[] { "--" + ShellFactory.FILE_LONG_OPTION, unrecognizedArg, tmpFile.getAbsolutePath() });

        assertTrue("file mode not detected", shell instanceof PetalsFileScriptShell);

        shell.run();

        final String error = systemErr.getLog();
        assertFalse(error.isEmpty());
        assertTrue(error.contains("FileNotFoundException"));
        assertTrue(error.contains(unrecognizedArg));

    }

    /**
     * Check the error management when Petals CLI is launched in file mode with
     * unknown argument after, using following command line:
     * <code>petals-cli --file existing-filename --unknown-arg</code>
     */
    @Test
    public void testnewShell_fileMode_error_04() throws Exception {
        final File tmpFile = File.createTempFile("petals-cli", "commands");
        final String unrecognizedArg = "--unknown-arg";

        try {
            this.getShell(new String[] { "--" + ShellFactory.FILE_LONG_OPTION,
                    tmpFile.getAbsolutePath(), unrecognizedArg });
            fail("UnrecognizedOptionException is not thrown");
        } catch (final UnrecognizedOptionException e) {
            // Expected exception
            assertTrue("Unrecognized argument is missing in the erreur message.", e
                    .getMessage().contains(unrecognizedArg));
        }
    }

    /**
     * Check that Petals CLI is launched in inlined mode using following command
     * line: <code>petals-cli -</code>
     */
    @Test
    public void testnewShell_inlinedMode() throws ParseException, ShellException, IOException,
            InvocationTargetException {

        final AbstractShell shell = this.getShell(new String[] { "-" });
        assertTrue("inlined mode not detected", shell instanceof PetalsInlinedScriptShell);
    }

}
